@@ -18,6 +18,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1335  USA
 
 *******************************************************/
 
+#include <my_global.h>
 #include <mysql_version.h>
 #include <my_base.h>
 #include <quicklz.h>
@@ -74,7 +75,6 @@ datasink_t datasink_compress = {
 	&compress_open,
 	&compress_write,
 	&compress_close,
-	&dummy_remove,
 	&compress_deinit
 };
 
@@ -369,7 +369,6 @@ create_worker_threads(uint n)
 				   thd)) {
 			msg("compress: pthread_create() failed: "
 			    "errno = %d", errno);
-			pthread_mutex_unlock(&thd->ctrl_mutex);
 			goto err;
 		}
 	}
@@ -386,13 +385,6 @@ create_worker_threads(uint n)
 	return threads;
 
 err:
-	while (i > 0) {
-		comp_thread_ctxt_t *thd;
-		i--;
-		thd = threads + i;
-		pthread_mutex_unlock(&thd->ctrl_mutex);
-	}
-
 	my_free(threads);
 	return NULL;
 }
