@@ -1,4 +1,4 @@
-/* Copyright (C) 2014 SkySQL Ab.
+/* Copyright (C) 2014, 2020, MariaDB
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA. */
 
-#include <my_global.h>
+#include "mariadb.h"
 #include <sql_class.h>
 #include <mysql/service_wsrep.h>
 
@@ -29,6 +29,15 @@ enum wsrep_conflict_state wsrep_thd_conflict_state(THD *, my_bool)
 int wsrep_is_wsrep_xid(const XID*)
 { return 0; }
 
+long long wsrep_xid_seqno(const XID* x)
+{ return -1; }
+
+const unsigned char* wsrep_xid_uuid(const XID*)
+{
+    static const unsigned char uuid[16] = {0};
+    return uuid;
+}
+
 bool wsrep_prepare_key(const uchar*, size_t, const uchar*, size_t, struct wsrep_buf*, size_t*)
 { return 0; }
 
@@ -148,5 +157,13 @@ void wsrep_unlock_rollback()
 void wsrep_set_data_home_dir(const char *)
 { }
 
+void wsrep_log(void (*)(const char *, ...), const char *, ...)
+{
+}
+
 my_bool wsrep_thd_is_applier(MYSQL_THD thd)
 { return false; }
+
+void wsrep_report_bf_lock_wait(MYSQL_THD thd,
+                               unsigned long long id)
+{}
