@@ -1065,7 +1065,7 @@ my_fill_mb2(CHARSET_INFO *cs, char *s, size_t slen, int fill)
   */
   for (last= s + slen - buflen; s <= last; s+= buflen)
   {
-    /* Enough space for the characer */
+    /* Enough space for the character */
     memcpy(s, buf, buflen);
   }
 
@@ -1160,12 +1160,9 @@ static size_t
 my_snprintf_mb2(CHARSET_INFO *cs __attribute__((unused)),
                 char* to, size_t n, const char* fmt, ...)
 {
-  size_t ret;
   va_list args;
   va_start(args,fmt);
-  ret= my_vsnprintf_mb2(to, n, fmt, args);
-  va_end(args);
-  return ret;
+  return my_vsnprintf_mb2(to, n, fmt, args);
 }
 
 
@@ -1230,27 +1227,27 @@ static inline int my_weight_mb2_utf16mb2_general_ci(uchar b0, uchar b1)
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)        my_weight_mb2_utf16mb2_general_ci(b0,b1)
 #define WEIGHT_MB4(b0,b1,b2,b3)  MY_CS_REPLACEMENT_CHARACTER
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #define MY_FUNCTION_NAME(x)      my_ ## x ## _utf16_bin
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)        ((int) MY_UTF16_WC2(b0, b1))
 #define WEIGHT_MB4(b0,b1,b2,b3)  ((int) MY_UTF16_WC4(b0, b1, b2, b3))
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #define DEFINE_STRNNCOLLSP_NOPAD
 #define MY_FUNCTION_NAME(x)      my_ ## x ## _utf16_general_nopad_ci
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)        my_weight_mb2_utf16mb2_general_ci(b0,b1)
 #define WEIGHT_MB4(b0,b1,b2,b3)  MY_CS_REPLACEMENT_CHARACTER
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #define DEFINE_STRNNCOLLSP_NOPAD
 #define MY_FUNCTION_NAME(x)      my_ ## x ## _utf16_nopad_bin
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)        ((int) MY_UTF16_WC2(b0, b1))
 #define WEIGHT_MB4(b0,b1,b2,b3)  ((int) MY_UTF16_WC4(b0, b1, b2, b3))
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #undef IS_MB2_CHAR
 #undef IS_MB4_CHAR
@@ -1456,7 +1453,7 @@ my_charlen_utf16(CHARSET_INFO *cs, const uchar *str, const uchar *end)
 #define MY_FUNCTION_NAME(x)       my_ ## x ## _utf16
 #define CHARLEN(cs,str,end)       my_charlen_utf16(cs,str,end)
 #define DEFINE_WELL_FORMED_CHAR_LENGTH_USING_CHARLEN
-#include "ctype-mb.inl"
+#include "ctype-mb.ic"
 #undef MY_FUNCTION_NAME
 #undef CHARLEN
 #undef DEFINE_WELL_FORMED_CHAR_LENGTH_USING_CHARLEN
@@ -1781,27 +1778,27 @@ struct charset_info_st my_charset_utf16_nopad_bin=
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)        my_weight_mb2_utf16mb2_general_ci(b1,b0)
 #define WEIGHT_MB4(b0,b1,b2,b3)  MY_CS_REPLACEMENT_CHARACTER
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #define MY_FUNCTION_NAME(x)      my_ ## x ## _utf16le_bin
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)        ((int) MY_UTF16_WC2(b1, b0))
 #define WEIGHT_MB4(b0,b1,b2,b3)  ((int) MY_UTF16_WC4(b1, b0, b3, b2))
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #define DEFINE_STRNNCOLLSP_NOPAD
 #define MY_FUNCTION_NAME(x)      my_ ## x ## _utf16le_general_nopad_ci
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)        my_weight_mb2_utf16mb2_general_ci(b1,b0)
 #define WEIGHT_MB4(b0,b1,b2,b3)  MY_CS_REPLACEMENT_CHARACTER
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #define DEFINE_STRNNCOLLSP_NOPAD
 #define MY_FUNCTION_NAME(x)      my_ ## x ## _utf16le_nopad_bin
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)        ((int) MY_UTF16_WC2(b1, b0))
 #define WEIGHT_MB4(b0,b1,b2,b3)  ((int) MY_UTF16_WC4(b1, b0, b3, b2))
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #undef IS_MB2_CHAR
 #undef IS_MB4_CHAR
@@ -2137,24 +2134,24 @@ static inline int my_weight_utf32_general_ci(uchar b0, uchar b1,
 #define MY_FUNCTION_NAME(x)      my_ ## x ## _utf32_general_ci
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB4(b0,b1,b2,b3)  my_weight_utf32_general_ci(b0, b1, b2, b3)
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #define MY_FUNCTION_NAME(x)      my_ ## x ## _utf32_bin
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB4(b0,b1,b2,b3)  ((int) MY_UTF32_WC4(b0, b1, b2, b3))
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #define DEFINE_STRNNCOLLSP_NOPAD
 #define MY_FUNCTION_NAME(x)      my_ ## x ## _utf32_general_nopad_ci
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB4(b0,b1,b2,b3)  my_weight_utf32_general_ci(b0, b1, b2, b3)
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #define DEFINE_STRNNCOLLSP_NOPAD
 #define MY_FUNCTION_NAME(x)      my_ ## x ## _utf32_nopad_bin
 #define WEIGHT_ILSEQ(x)          (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB4(b0,b1,b2,b3)  ((int) MY_UTF32_WC4(b0, b1, b2, b3))
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 #undef IS_MB2_CHAR
 #undef IS_MB4_CHAR
@@ -2327,7 +2324,7 @@ my_charlen_utf32(CHARSET_INFO *cs __attribute__((unused)),
 #define MY_FUNCTION_NAME(x)       my_ ## x ## _utf32
 #define CHARLEN(cs,str,end)       my_charlen_utf32(cs,str,end)
 #define DEFINE_WELL_FORMED_CHAR_LENGTH_USING_CHARLEN
-#include "ctype-mb.inl"
+#include "ctype-mb.ic"
 #undef MY_FUNCTION_NAME
 #undef CHARLEN
 #undef DEFINE_WELL_FORMED_CHAR_LENGTH_USING_CHARLEN
@@ -2427,12 +2424,9 @@ static size_t
 my_snprintf_utf32(CHARSET_INFO *cs __attribute__((unused)),
                   char* to, size_t n, const char* fmt, ...)
 {
-  size_t ret;
   va_list args;
   va_start(args,fmt);
-  ret= my_vsnprintf_utf32(to, n, fmt, args);
-  va_end(args);
-  return ret;
+  return my_vsnprintf_utf32(to, n, fmt, args);
 }
 
 
@@ -2628,14 +2622,14 @@ void my_fill_utf32(CHARSET_INFO *cs,
                    char *s, size_t slen, int fill)
 {
   char buf[10];
-#ifndef DBUG_OFF
+#ifdef DBUG_ASSERT_EXISTS
   uint buflen;
 #endif
   char *e= s + slen;
   
   DBUG_ASSERT((slen % 4) == 0);
 
-#ifndef DBUG_OFF
+#ifdef DBUG_ASSERT_EXISTS
   buflen=
 #endif
     cs->cset->wc_mb(cs, (my_wc_t) fill, (uchar*) buf,
@@ -2993,7 +2987,7 @@ static const uchar to_upper_ucs2[] = {
 };
 
 
-/* Definitions for strcoll.inl */
+/* Definitions for strcoll.ic */
 #define IS_MB2_CHAR(x,y)            (1)
 #define UCS2_CODE(b0,b1)            (((uchar) b0) << 8 | ((uchar) b1))
 
@@ -3009,27 +3003,27 @@ static inline int my_weight_mb2_ucs2_general_ci(uchar b0, uchar b1)
 #define MY_FUNCTION_NAME(x)    my_ ## x ## _ucs2_general_ci
 #define WEIGHT_ILSEQ(x)        (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)      my_weight_mb2_ucs2_general_ci(b0,b1)
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 
 #define MY_FUNCTION_NAME(x)    my_ ## x ## _ucs2_bin
 #define WEIGHT_ILSEQ(x)        (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)      UCS2_CODE(b0,b1)
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 
 #define DEFINE_STRNNCOLLSP_NOPAD
 #define MY_FUNCTION_NAME(x)    my_ ## x ## _ucs2_general_nopad_ci
 #define WEIGHT_ILSEQ(x)        (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)      my_weight_mb2_ucs2_general_ci(b0,b1)
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 
 #define DEFINE_STRNNCOLLSP_NOPAD
 #define MY_FUNCTION_NAME(x)    my_ ## x ## _ucs2_nopad_bin
 #define WEIGHT_ILSEQ(x)        (0xFF0000 + (uchar) (x))
 #define WEIGHT_MB2(b0,b1)      UCS2_CODE(b0,b1)
-#include "strcoll.inl"
+#include "strcoll.ic"
 
 
 static int
